package ch.unisg.tapas.auctionhouse.adapter.common.clients;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.http.HttpClient;
import java.net.http.HttpRequest;
import java.net.http.HttpResponse;
import java.util.HashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.json.JSONObject;
import org.springframework.http.HttpStatus;

/**
 * Subscribes to the WebSub hubs of auction houses discovered at run time. This class is instantiated
 * from {@link ch.unisg.tapas.TapasAuctionHouseApplication} when boostraping the TAPAS marketplace
 * via WebSub.
 */
public class WebSubSubscriber {

    Logger logger = Logger.getLogger(WebSubSubscriber.class.getName());

    String AUCTION_HOUSE_ENDPOINT;

    public WebSubSubscriber(String AUCTION_HOUSE_ENDPOINT) {
        this.AUCTION_HOUSE_ENDPOINT = AUCTION_HOUSE_ENDPOINT;
    }

    public void subscribeToAuctionHouseEndpoint(URI endpoint) {

        HashMap<String, String> links = discoverWebSubTopic(endpoint);

        if (links.isEmpty()) {
            return;
        }

        subscribeToWebSub(links.get("hub"), links.get("self"));

        // 1. Send a request to the auction house in order to discover the WebSub hub to subscribe to.
        // The request URI should depend on the design of the Auction House HTTP API.
        // 2. Send a subscription request to the discovered WebSub hub to subscribe to events relevant
        // for this auction house.
        // 3. Handle the validation of intent from the WebSub hub (see WebSub protocol).
        //
        // Once the subscription is activated, the hub will send "fat pings" with content updates.
        // The content received from the hub will depend primarily on the design of the Auction House
        // HTTP API.
        //
        // For further details see:
        // - W3C WebSub Recommendation: https://www.w3.org/TR/websub/
        // - the implementation notes of the WebSub hub you are using to distribute events
    }

    private HashMap<String, String> discoverWebSubTopic(URI endpoint) {
        HttpClient client = HttpClient.newHttpClient();
        HttpRequest request = HttpRequest.newBuilder()
                .uri(endpoint)
                .header("Content-Type", "application/json")
                .GET()
                .build();

        HashMap<String, String> links = new HashMap<>();

        try {
            HttpResponse<String> response = client.send(request, HttpResponse.BodyHandlers.ofString());
            if (response.statusCode() == HttpStatus.OK.value()) {
                for (String link : response.headers().allValues("link")) {
                    if (link.contains("rel=\"hub\"")) {
                        String hub = link.split(">")[0];
                        links.put("hub", hub.substring(1));
                    } else if(link.contains("rel=\"self\"")) {
                        String self = link.split(">")[0];
                        links.put("self", self.substring(1));
                    }
                    System.out.println(link);
                }
                // TODO check for HTML tags second if links are not present in headers
            } else {
                logger.log(Level.SEVERE, "Could not find a websub uri");
            }
        } catch (InterruptedException e) {
            logger.log(Level.SEVERE, e.getLocalizedMessage(), e);
            Thread.currentThread().interrupt();
        } catch (IOException e) {
            logger.log(Level.SEVERE, e.getLocalizedMessage(), e);
        }
        return links;
    }

    private void subscribeToWebSub(String hub, String topic) {
        HttpClient client = HttpClient.newHttpClient();

        URI hubURI;
        try {
            hubURI = new URI(hub);
        } catch (URISyntaxException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            return;
        }

        System.out.println("HUB: " + hub);
        System.out.println("TOPIC: " + topic);

        HttpRequest request = HttpRequest.newBuilder()
                .uri(hubURI)
                .header("Content-Type", "application/x-www-form-urlencoded")
                .POST(HttpRequest.BodyPublishers.ofString("hub.mode=subscribe&hub.callback=" + AUCTION_HOUSE_ENDPOINT +
                "/auction-started/74c72c7f-2739-4124-aa35-a3225171a97c" + "&hub.topic=" + topic))
                .build();


        try {
            client.send(request, HttpResponse.BodyHandlers.ofString());
        } catch (InterruptedException e) {
            logger.log(Level.SEVERE, e.getLocalizedMessage(), e);
            Thread.currentThread().interrupt();
        } catch (IOException e) {
            logger.log(Level.SEVERE, e.getLocalizedMessage(), e);
        }
    }
}
